#include "io/FontAtlas.h"
#include "io/File.h"
#include "Serializer.h"

namespace ehs
{
	FontAtlas::FontAtlas()
		: glyphScale(0)
	{
	}

	FontAtlas::FontAtlas(const Str_8& filePath)
		: glyphScale(0)
	{
		File fontFile(filePath, Mode::READ, Disposition::OPEN);

		hashId = fontFile.GetName().Hash_64();
		id = fontFile.GetName();

		Serializer<UInt_64> fData = fontFile.ReadSerializer_64(Endianness::LE, fontFile.Size());
		fontFile.Release();

		Version ver = fData.ReadVersion();
		if (ver != Version(1, 0, 0))
		{
			EHS_LOG_INT("Error", 2, "The Event Horizon Font file, \"" + filePath + "\", must be version 1.0.0, but was version " +
									Str_8::FromNum(ver.major) + "." + Str_8::FromNum(ver.minor) + "." +
									Str_8::FromNum(ver.patch) + ".");
			return;
		}

		glyphScale = fData.Read<UInt_64>();

		glyphs.Resize(fData.Read<UInt_64>());
		for (UInt_64 i = 0; i < glyphs.Size(); ++i)
			glyphs[i] = Glyph(fData);

		bitDepth = 8;
		channels = 1;
		width = fData.Read<UInt_64>();
		height = fData.Read<UInt_64>();
		data = new Byte[width * height * (bitDepth / 8) * channels];
		fData.ReadArray(data, &size);
	}

	FontAtlas::FontAtlas(FontAtlas&& fa) noexcept
		: Img(std::move(fa)), glyphScale(fa.glyphScale), glyphs(std::move(fa.glyphs))
	{
		fa.glyphScale = 0;
	}

	FontAtlas::FontAtlas(const FontAtlas& fa)
		: Img(fa), glyphScale(0)
	{
	}

	FontAtlas& FontAtlas::operator=(FontAtlas&& fa) noexcept
	{
		if (this == &fa)
			return *this;

		Img::operator=(std::move(fa));

		glyphScale = fa.glyphScale;
		glyphs = std::move(fa.glyphs);

		fa.glyphScale = 0;

		return *this;
	}

	FontAtlas& FontAtlas::operator=(const FontAtlas& fa)
	{
		if (this == &fa)
			return *this;

		Img::operator=(fa);

		glyphScale = 0;
		glyphs = Array<Glyph>();

		return *this;
	}

	UInt_64 FontAtlas::GetGlyphScale() const
	{
		return glyphScale;
	}

	Glyph FontAtlas::GetGlyph(const Char_32 code) const
	{
		for (UInt_32 i = 0; i < glyphs.Size(); ++i)
			if (glyphs[i].GetCode() == code)
				return glyphs[i];

		return glyphs[0];
	}

	Vec2_f FontAtlas::CalculateSize(const Str_8& text) const
	{
		Vec2_f size;

		for (UInt_64 i = 0; i < text.Size(); ++i)
		{
			Glyph glyph = GetGlyph(text[i]);

			size.x += (float)glyph.GetAdvance().x;

			if ((float)glyph.GetScale().y > size.y)
				size.y = (float)glyph.GetScale().y;
		}

		return size;
	}

	float FontAtlas::CalculateWidth(const Str_8 &text) const
	{
		float width = 0.0f;

		for (UInt_64 i = 0; i < text.Size(); ++i)
		{
			Glyph glyph = GetGlyph(text[i]);

			width += (float)glyph.GetAdvance().x;
		}

		return width;
	}

	float FontAtlas::CalculateHeight(const Str_8& text) const
	{
		float height = 0.0f;

		for (UInt_64 i = 0; i < text.Size(); ++i)
		{
			Glyph glyph = GetGlyph(text[i]);

			if ((float)glyph.GetScale().y > height)
				height = (float)glyph.GetScale().y;
		}

		return height;
	}

	UInt_64 FontAtlas::CalculateIndexAtPoint(const Str_8& text, const Vec2_f& point) const
	{
		float width = 0.0f;

		for (UInt_64 i = 0; i < text.Size(); ++i)
		{
			Glyph glyph = GetGlyph(text[i]);

			float temp = width + (float)glyph.GetAdvance().x;
			if (point.x > temp)
				width += (float)glyph.GetAdvance().x;
			else if (point.x <= temp)
				return i;
		}

		return text.Size();
	}

	Mesh FontAtlas::Generate(const Anchor anchor, const Str_8& text) const
	{
		Vec2_f pos;

		switch (anchor)
		{
			case Anchor::BOTTOM_CENTER:
			{
				pos.x -= CalculateWidth(text) * 0.5f;
				break;
			}
			case Anchor::TOP_LEFT:
			{
				pos.x -= CalculateWidth(text);
				break;
			}
			case Anchor::TOP_CENTER:
			{
				pos.x -= CalculateWidth(text) * 0.5f;
				break;
			}
			case Anchor::CENTER_RIGHT:
			{
				pos.y -= CalculateHeight(text);
				break;
			}
			case Anchor::CENTER:
			{
				Vec4_f size = CalculateSize(text);
				pos.x -= size.x * 0.5f;
				pos.y -= size.y;
				break;
			}
		}

		Array<Vertex_f> verts(text.Size() * 4);
		Array<UInt_32> indices(text.Size() * 6);

		for (UInt_64 i = 0; i < text.Size(); ++i)
		{
			Glyph glyph = GetGlyph(text[i]);
			Vec2_f scale = glyph.GetScale();

			float x = pos.x + (float)glyph.GetBearing().x;
			float y = pos.y + (float)(GetGlyphScale() - glyph.GetBearing().y);

			UInt_32 vertsI = i * 4;

			verts[vertsI] = Vertex_f({x, y, 1.0f}, {}, glyph.GetUV().GetPos());
			verts[vertsI + 1] = Vertex_f({x, y + scale.y, 1.0f}, {}, Vec2_f(glyph.GetUV().x, glyph.GetUV().h));
			verts[vertsI + 2] = Vertex_f({x + scale.x, y, 1.0f}, {}, Vec2_f(glyph.GetUV().w, glyph.GetUV().y));
			verts[vertsI + 3] = Vertex_f({x + scale.x, y + scale.y, 1.0f}, {}, glyph.GetUV().GetScale());

			UInt_32 indicesI = i * 6;

			indices[indicesI] = vertsI;
			indices[indicesI + 1] = vertsI + 1;
			indices[indicesI + 2] = vertsI + 2;
			indices[indicesI + 3] = vertsI + 3;
			indices[indicesI + 4] = vertsI + 2;
			indices[indicesI + 5] = vertsI + 1;

			pos.x += (float)glyph.GetAdvance().x;
		}

		return {id, std::move(verts), std::move(indices)};
	}
}